
#include <globdev.h>
#include "mpid.h"
#include "mpiimpl.h"
#include "protos.h"
#include "attr.h"

#include "topology_intra_fns.h"


/**********************************************************************/
/* GLOBAL SYMBOLS                                                     */
/**********************************************************************/

/* key to retrieve the Topology information from each communicator */
int MPICHX_TOPOLOGY_COLORS = MPI_KEYVAL_INVALID;
int MPICHX_TOPOLOGY_DEPTHS = MPI_KEYVAL_INVALID;
MPI_Delete_function mpichx_topology_depths_destructor;
MPI_Delete_function mpichx_topology_colors_destructor;



/**********************************************************************/
/* PRIVATE FUNCTIONS                                                  */
/**********************************************************************/


/**********************************************************************/
/* allocate memory and check the return pointer.  MPID_Abort if NULL */
#define g_malloc_chk(Ptr, Type, Size) \
{ \
   const size_t sz = (Size); \
   if ( sz > 0 ) \
   { \
      Ptr = (Type) globus_libc_malloc (sz); \
      if ( !Ptr ) \
      { \
         int me, size; \
         MPI_Comm_rank(MPI_COMM_WORLD, &me); \
         MPI_Comm_size(MPI_COMM_WORLD, &size); \
         globus_libc_fprintf(stderr, "[%d/%d:%s:%d] failed malloc %d bytes\n", \
                             me, size, __FILE__, __LINE__, (int)sz); \
         MPID_Abort((struct MPIR_COMMUNICATOR*) 0, 2, "MPICH-G2 Internal", \
                    __FILE__ " - failed malloc"); \
      } \
   } \
   else Ptr = (Type) NULL; \
}


/**********************************************************************/
/* dump the topology information attached to a given communicator */
static void
print_topology (struct MPIR_COMMUNICATOR * const comm)
{
   MPICHX_Topolevel max_depth = 0;
   int i, j, rank, size;

   (void) MPIR_Comm_size(comm, &size);
   (void) MPIR_Comm_rank(comm, &rank);

   globus_libc_fprintf(stderr, "*** Start print topology from proc #%d/%d\n",
                       rank, size);
   globus_libc_fprintf(stderr, "Sizes of my clusters:\n");
   for (i = 0; i < comm->Topology_Depths[rank]; i++)
      globus_libc_fprintf(stderr, "Level %d: %d procs\n", i,
                          comm->Topology_ClusterSizes[i]);
   globus_libc_fprintf(stderr, "proc\t");
   for (i = 0; i < size; i++)
      globus_libc_fprintf(stderr, "% 3d", i);
   globus_libc_fprintf(stderr, "\ndepths\t");
   for (i = 0; i < size; i++)
   {
      if ( max_depth < comm->Topology_Depths[i] )
         max_depth = comm->Topology_Depths[i];
      globus_libc_fprintf(stderr, "% 3d", comm->Topology_Depths[i]);
   }
   globus_libc_fprintf(stderr, "\nCOLORS:");
   for (i = 0; i < max_depth; i++)
   {
      globus_libc_fprintf(stderr, "\nlvl %d\t", i);
      for (j = 0; j < size; j++)
         if ( i < comm->Topology_Depths[j] )
            globus_libc_fprintf(stderr, "% 3d", comm->Topology_Colors[j][i]);
         else
            globus_libc_fprintf(stderr, "   ");
   }
   globus_libc_fprintf(stderr, "\nCLUSTER_IDS:");
   for (i = 0; i < max_depth; i++)
   {
      globus_libc_fprintf(stderr, "\nlvl %d\t", i);
      for (j = 0; j < size; j++)
         if ( i < comm->Topology_Depths[j] )
            globus_libc_fprintf(stderr, "% 3d",
                                comm->Topology_ClusterIds[j][i]);
         else
            globus_libc_fprintf(stderr, "   ");
   }
   globus_libc_fprintf(stderr, "\n");
   globus_libc_fprintf(stderr, "*** End print topology\n");
   return;
}


/**********************************************************************/
/* return the number of protocol levels thru which a process can
 * communicate */
static MPICHX_Topolevel
num_protos_in_channel (struct channel_t * const cp)
{
   struct miproto_t *mp;
   MPICHX_Topolevel rc = 0;

   if (!cp)
   {
      globus_libc_fprintf(stderr,
               "\tERROR: num_protos_in_channel(): grank %d: passed NULL cp\n",
               MPID_MyWorldRank);
      MPID_Abort( (struct MPIR_COMMUNICATOR *)0, 1, "MPICH-G2", "");
   } /* end if */

   for (mp = cp->proto_list; mp; mp = mp->next)
   {
      switch (mp->type)
      {
         /* TCP: 1 for localhost + 1 for LAN + 1 for WAN */
         case tcp: rc += 3; break;
         case mpi: rc ++; break;
         default: globus_libc_fprintf(stderr,
                     "\tERROR: num_protos_in_channel(): grank %d: encountered "
                     "unrecognized proto type %d", MPID_MyWorldRank, mp->type);
                  MPID_Abort( (struct MPIR_COMMUNICATOR *)0, 1, "MPICH-G2", "");
      } /* end switch */
   } /* end for */

   return rc;
} /* end num_protos_in_channel() */


/**********************************************************************/
/* return TRUE if the two processes can talk to each other at the
 * given level; FALSE otherwise */
static int
channels_proto_match (struct channel_t * const cp0,
                      struct channel_t * const cp1,
                      const MPICHX_Topolevel level)
{
   struct miproto_t *mp0, *mp1;
   enum proto type;

   /* if level == MPICHX_WAN_LEVEL, then they always match */
   if ( level == MPICHX_WAN_LEVEL ) return GLOBUS_TRUE;

   switch ( level )
   {
      /* level != MPICHX_WAN_LEVEL */
      case MPICHX_LAN_LEVEL:
      case MPICHX_HOST_LEVEL: type = tcp; break;
      case MPICHX_VMPI_LEVEL: type = mpi; break;
      default:
         MPID_Abort((struct MPIR_COMMUNICATOR *)0, 1, "MPICH-G2 Internal",
                    "channels_proto_match(): unrecognized topology level");
   }

   /* finding the correct protos in each channel */
   for (mp0 = cp0->proto_list; mp0; mp0 = mp0->next)
      if ( mp0->type == type ) break;
   for (mp1 = cp1->proto_list; mp1; mp1 = mp1->next)
      if ( mp1->type == type ) break;

   if ( mp0  &&  mp1  &&  (mp0->type == mp1->type) )
   {
      /* now that i have correct proto for each, seeing if they match */
      switch ( level )
      {
         /* level != MPICHX_WAN_LEVEL */
         case MPICHX_LAN_LEVEL:   /* are the procs in the same LAN? */
         {
            if ( !strcmp(((struct tcp_miproto_t *)(mp0->info))->globus_lan_id,
                         ((struct tcp_miproto_t *)(mp1->info))->globus_lan_id) )
               return GLOBUS_TRUE;
            else
               return GLOBUS_FALSE;
         }
         case MPICHX_HOST_LEVEL:/* are the procs on the same localhost? */
         {
            if ( ((struct tcp_miproto_t *) (mp0->info))->localhost_id ==
                 ((struct tcp_miproto_t *) (mp1->info))->localhost_id )
               return GLOBUS_TRUE;
            else
               return GLOBUS_FALSE;
         }
         case MPICHX_VMPI_LEVEL:
         {
            if ( !strcmp( ((struct mpi_miproto_t *) 
                                        (mp0->info))->unique_session_string,
                          ((struct mpi_miproto_t *) 
                                        (mp1->info))->unique_session_string ) )
               return GLOBUS_TRUE;
            else
               return GLOBUS_FALSE;
         }
         default:
            MPID_Abort((struct MPIR_COMMUNICATOR *)0, 1, "MPICH-G2 Internal",
                  "channels_proto_match(): unrecognized topology level");
      } /* end switch */
   } /* end if */

   return GLOBUS_FALSE;
} /* end channels_proto_match() */


/**********************************************************************/
/* create the keys for the attributes which allows the user to access
 * the underlying topology of processes; register the attributes
 * (depths and colors) into the communicator */
static int
create_topology_keys(struct MPIR_COMMUNICATOR * const comm)
{
   int mpi_errno = MPI_SUCCESS;
   int flag, size;
   MPICHX_Topolevel *Depths;
   int **Colors;

   (void) MPIR_Comm_size (comm, &size);

   if ( MPICHX_TOPOLOGY_DEPTHS == MPI_KEYVAL_INVALID )
   {
      mpi_errno = MPI_Keyval_create(MPI_NULL_COPY_FN,
                                    mpichx_topology_depths_destructor,
                                    &MPICHX_TOPOLOGY_DEPTHS, NULL);
      if ( mpi_errno ) return mpi_errno;
   }

   mpi_errno = MPI_Attr_get(comm->self, MPICHX_TOPOLOGY_DEPTHS, &Depths,
                            &flag);
   if ( mpi_errno ) return mpi_errno;

   if ( !flag )
   {
      int i;

      /* copy the information available at the user level: the user must
       * not have a direct access to the pointer to the real data used by
       * the MPICH library */
      g_malloc_chk(Depths, MPICHX_Topolevel *, sizeof(MPICHX_Topolevel) * size);

      for (i = 0; i < size; i++)
         Depths[i] = comm->Topology_Depths[i];
      mpi_errno = MPI_Attr_put(comm->self, MPICHX_TOPOLOGY_DEPTHS, Depths);
      if ( mpi_errno ) return mpi_errno;
   }

   if ( MPICHX_TOPOLOGY_COLORS == MPI_KEYVAL_INVALID )
   {
      mpi_errno = MPI_Keyval_create(MPI_NULL_COPY_FN,
                                    mpichx_topology_colors_destructor,
                                    &MPICHX_TOPOLOGY_COLORS, NULL);
      if ( mpi_errno ) return mpi_errno;
   }

   mpi_errno = MPI_Attr_get(comm->self, MPICHX_TOPOLOGY_COLORS, &Colors,
                            &flag);
   if ( mpi_errno ) return mpi_errno;

   if ( !flag )
   {
      int i;

      /* copy the information available at the user level: the user must
       * not have a direct access to the pointer to the real data used by
       * the MPICH library */
      g_malloc_chk(Colors, int **, sizeof(int *) * size);

      for (i = 0; i < size; i++)
      {
         int j;
         const MPICHX_Topolevel depth = Depths[i];
         int *column = comm->Topology_Colors[i];

         g_malloc_chk(Colors[i], int *, sizeof(int) * depth);
         for (j = 0; j < depth; j++)
            Colors[i][j] = column[j];
      }

      mpi_errno = MPI_Attr_put(comm->self, MPICHX_TOPOLOGY_COLORS, Colors);
      if ( mpi_errno ) return mpi_errno;
   }

   return mpi_errno;
}



/**********************************************************************/
/* PUBLIC FUNCTIONS                                                   */
/**********************************************************************/

/**********************************************************************/
/* initialize Topology_Depths, Topology_Colors and Topology_ClusterIds
 * for topology-aware collective operations and topology reporting to
 * MPI app */
int
topology_initialization (struct MPIR_COMMUNICATOR * const comm)
{
   MPICHX_Topolevel lvl, my_depth, max_depth;
   int rank, size, p0;
   int mpi_errno = MPI_SUCCESS;
   MPICHX_Topolevel *Depths;
   int **ClusterIds, **Colors, *ClusterSizes;
   comm_set_t *CommSets;

   /* don't do anything for intercommunicators */
   if (comm->comm_type == MPIR_INTER) return mpi_errno;

   (void) MPIR_Comm_rank(comm, &rank);
   (void) MPIR_Comm_size(comm, &size);

   /* allocate memory for the array of depths */
   g_malloc_chk(Depths, MPICHX_Topolevel *, size * sizeof(MPICHX_Topolevel));
   comm->Topology_Depths = Depths;

   /* allocate memory for the 2D-array of colors */
   g_malloc_chk(Colors, int **, size * sizeof(int *));
   comm->Topology_Colors = Colors;

   /* allocate memory for the 2D-array of cluster IDs */
   g_malloc_chk(ClusterIds, int **, size * sizeof(int *));
   comm->Topology_ClusterIds = ClusterIds;


   /*********************************/
   /* Phase 1 of 3 - Finding Depths */
   /*********************************/

   max_depth = 0;
   for (p0 = 0; p0 < size; p0++)
   {
      struct channel_t *chanl;
      MPICHX_Topolevel depth;

      chanl = get_channel(comm->lrank_to_grank[p0]);
      if ( !chanl )
         MPID_Abort((struct MPIR_COMMUNICATOR *)0, 2, "MPICH-G2 Internal",
                    "topology_initialization() - NULL channel returned");
      Depths[p0] = depth = num_protos_in_channel(chanl);

      if (depth > max_depth) max_depth = depth;

      /* allocate memory for the colors and cluster-IDs */
      g_malloc_chk(Colors[p0], int *, depth * sizeof(int));
      g_malloc_chk(ClusterIds[p0], int *, depth * sizeof(int));

      /* initialize the colors and cluster IDs to invalid value -1 */
      for (lvl = 0; lvl < depth; lvl ++)
      {
         Colors[p0][lvl] = -1;
         ClusterIds[p0][lvl] = -1;
      }
   } /* endfor */
   my_depth = Depths[rank];

   /* allocate memory for the sets of communicating processes I will
    * be involved in */
   g_malloc_chk(CommSets, comm_set_t *, my_depth * sizeof(comm_set_t));
   comm->Topology_CommSets = CommSets;

   /* allocate memory for the sizes of the clusters I belong to at
    * each communication level */
   g_malloc_chk(ClusterSizes, int *, sizeof(int) * my_depth);
   comm->Topology_ClusterSizes = ClusterSizes;


   /***************************/
   /* Phase 2 of 3 - Coloring */
   /***************************/

   for (lvl = 0; lvl < max_depth; lvl ++)
   {
      int next_color = 0;

      for (p0 = 0; p0 < size; p0++)
      {
         if (lvl < Depths[p0] && Colors[p0][lvl] == -1)
         {
            /* this proc has not been colored at this level yet, i.e.,
             * it hasn't matched any of the procs to the left at this
             * level yet ... ok, start new color at this level. */

            struct channel_t *chanl0;
            int p1, color0;

            chanl0 = get_channel(comm->lrank_to_grank[p0]);
            if ( !chanl0 )
               MPID_Abort((struct MPIR_COMMUNICATOR *)0, 2, "MPICH-G2 Internal",
                          "topology_initialization() - NULL channel returned");

            Colors[p0][lvl] = color0 = next_color ++;
            for (p1 = p0 + 1; p1 < size; p1++)
            {
               if (lvl < Depths[p1] && Colors[p1][lvl] == -1)
               {
                  struct channel_t *chanl1;;

                  chanl1 = get_channel(comm->lrank_to_grank[p1]);
                  if ( !chanl1 )
                     MPID_Abort((struct MPIR_COMMUNICATOR *)0, 2,
                                "MPICH-G2 Internal",
                                "topology_initialization() - NULL channel");

                  if (channels_proto_match(chanl0, chanl1, lvl))
                     Colors[p1][lvl] = color0;
               } /* endif */
            } /* endfor */
         } /* endif */
      } /* endfor */
   } /* endfor */


   /************************************************/
   /* Phase 3 of 3 - Setting CID's based on colors */
   /************************************************/

   for (lvl = max_depth-1; lvl >= 0; lvl--)
   {
      for (p0 = 0; p0 < size; p0 ++)
      {
         if (lvl < Depths[p0] && ClusterIds[p0][lvl] == -1)
         {
            /* p0 has not been assigned a cid at this level yet, which
             * means all the procs at this level that have the same
             * color as p0 at this level have also not been assigned
             * cid's yet.
             *
             * find the rest of the procs at this level that have the
             * same color as p0 and assign them cids.  same color are
             * enumerated.  */

            int next_cid, p1;
            const int color0 = Colors[p0][lvl];

            ClusterIds[p0][lvl] = 0;
            next_cid = 1;

            for (p1 = p0 + 1; p1 < size; p1++)
            {
               if ( lvl < Depths[p1]  &&  color0 == Colors[p1][lvl] )
               {
                  /* p0 and p1 match colors at this level, which means
                   * p1 will now get its cid set at this level.  but
                   * to what value?  if p1 also matches color with any
                   * proc to its left at level lvl+1, then p1 copies
                   * that proc's cid at this level, otherwise p1 gets
                   * the next cid value at this level.  */

                  if (lvl+1 < Depths[p1])
                  {
                     int p2;
                     const int color1 = Colors[p1][lvl];
                     const int next_color1 = Colors[p1][lvl+1];

                     for (p2 = 0; p2 < p1; p2 ++)
                     {
                        if ( lvl+1 < Depths[p2] &&
                             color1 == Colors[p2][lvl] &&
                             next_color1 == Colors[p2][lvl+1])
                        {
                           ClusterIds[p1][lvl] = ClusterIds[p2][lvl];
                           break;
                        }
                     } /* endfor */
                     if (p2 == p1)
                        /* did not find one */
                        ClusterIds[p1][lvl] = next_cid ++;
                  }
                  else
                     /* p1 does not have a level lvl+1 */
                     ClusterIds[p1][lvl] = next_cid ++;
               } /* endif */
            } /* endfor */
         } /* endif */
      } /* endfor */
   } /* endfor */

   /* know how many processes are in my cluster at each level and
    * allocate enough memory for the sets of communicating processes
    * at each level */
   for (lvl = 0; lvl < my_depth; lvl++)
   {
      /* number of procs in the same cluster as I: useful to allocate
       * enough memory to store the data i may have to forward to
       * other processes in my cluster */
      int cnt = 0;
      /* number of procs I may have to talk to at this level */
      int n_cid = 0;
      int my_color = Colors[rank][lvl];

      for (p0 = 0; p0 < size; p0++)
         if ( Depths[p0] > lvl  &&  my_color == Colors[p0][lvl] )
         {
            cnt++;
            if ( ClusterIds[p0][lvl] > n_cid )
               n_cid = ClusterIds[p0][lvl];
         }
      comm->Topology_ClusterSizes[lvl] = cnt;
      g_malloc_chk(CommSets[lvl].set, int*, sizeof(int) * (n_cid + 1));
   }

   mpi_errno = create_topology_keys(comm);

/*
if ( size > 1  &&  rank == 1 ) print_topology(comm);
*/

   return mpi_errno;
}


/**********************************************************************/
/* free memory used for the topology information attached to the given
 * communicator */
void
topology_destruction (struct MPIR_COMMUNICATOR * const comm)
{
   int i, rank, size;
   MPICHX_Topolevel my_depth;

   (void) MPIR_Comm_rank(comm, &rank);
   (void) MPIR_Comm_size(comm, &size);

   /* don't do anything for intercommunicators */
   if (comm->comm_type == MPIR_INTER) return;

   for (i = 0; i < size; i ++)
   {
      g_free(comm->Topology_Colors[i]);
      g_free(comm->Topology_ClusterIds[i]);
   }
   g_free(comm->Topology_Colors);
   comm->Topology_Colors = NULL;
   g_free(comm->Topology_ClusterIds);
   comm->Topology_ClusterIds = NULL;

   my_depth = comm->Topology_Depths[rank];
   g_free(comm->Topology_Depths);
   comm->Topology_Depths = NULL;

   for (i = 0; i < my_depth; i ++)
      g_free(comm->Topology_CommSets[i].set);
   g_free(comm->Topology_CommSets);
   comm->Topology_CommSets = NULL;

   g_free(comm->Topology_ClusterSizes);
   comm->Topology_ClusterSizes = NULL;

   return;
}


/**********************************************************************/
/* this function is called as a communicator is destroyed: free memory
 * used for the array of Topology_Depths */
int
mpichx_topology_depths_destructor (MPI_Comm comm, int key, void *attr,
                                   void *extra)
{
   int mpi_errno = MPI_SUCCESS;
   MPICHX_Topolevel *Depths;

   /* if the key is not valid any longer, then the attribute has already
    * been deleted (and memory freed) */
   if ( MPICHX_TOPOLOGY_DEPTHS == MPI_KEYVAL_INVALID )
      return mpi_errno;

   if ( key != MPICHX_TOPOLOGY_DEPTHS )
      /* what does that mean?  What return value?  If there's no bug in MPICH
       * library, this case should never be encountered... */
      return mpi_errno;

   Depths = (MPICHX_Topolevel *) attr;
   g_free(Depths);

   return mpi_errno;
}


/**********************************************************************/
/* this function is called as a communicator is destroyed: free memory
 * used for the 2D-array of Topology_Colors */
int
mpichx_topology_colors_destructor (MPI_Comm comm, int key, void *attr,
                                   void *extra)
{
   int mpi_errno = MPI_SUCCESS;
   int i, size;
   int **Colors;

   /* if the key is not valid any longer, then the attribute has already
    * been deleted (and memory freed) */
   if ( MPICHX_TOPOLOGY_COLORS == MPI_KEYVAL_INVALID )
      return mpi_errno;

   if ( key != MPICHX_TOPOLOGY_COLORS )
      /* what does that mean?  What return value?  If there's no bug in MPICH
       * library, this case should never be encountered... */
      return mpi_errno;

   Colors = (int **) attr;
   (void) MPI_Comm_size(comm, &size);

   for (i = 0; i < size; i++)
      g_free(Colors[i]);
   g_free(Colors);

   return mpi_errno;
}

