
#include "mpid.h"
#include "mpiimpl.h"
#include "mem.h"
#include "coll.h"
#include "mpiops.h"

#include "topology_intra_fns.h"



/**********************************************************************/
/* the library routines should call the profiling versions of the
 * functions (PMPI_...) to ensure that only any user code to catch an
 * MPI function only uses the PMPI functions.  Here is a partial
 * solution.  In the weak symbol case, we simply change all of the
 * routines to their PMPI versions. */
/**********************************************************************/

/* Include mapping from MPI->PMPI */
#define MPI_BUILD_PROFILING
#include "mpiprof.h"
/* Insert the prototypes for the PMPI routines */
#undef __MPI_BINDINGS
#include "binding.h"



/**********************************************************************/
/* PRIVATE FUNCTIONS                                                  */
/**********************************************************************/

/**********************************************************************/
/* allocate memory and check the return pointer.  MPID_Abort if NULL */
#define g_malloc_chk(Ptr, Type, Size) \
{ \
   const size_t sz = (Size); \
   if ( sz > 0 ) \
   { \
      Ptr = (Type) globus_libc_malloc (sz); \
      if ( !Ptr ) \
      { \
         int me, size; \
         MPI_Comm_rank(MPI_COMM_WORLD, &me); \
         MPI_Comm_size(MPI_COMM_WORLD, &size); \
         globus_libc_fprintf(stderr, "[%d/%d:%s:%d] failed malloc %d bytes\n", \
                             me, size, __FILE__, __LINE__, (int)sz); \
         MPID_Abort((struct MPIR_COMMUNICATOR*) 0, 2, "MPICH-G2 Internal", \
                    __FILE__ " - failed malloc"); \
      } \
   } \
   else Ptr = (Type) NULL; \
}


/**********************************************************************/
/* print the contents of the sets of communicating processes attached
 * to the given communicator */
static void
print_comm_set (struct MPIR_COMMUNICATOR * const comm)
{
   MPICHX_Topolevel lvl, my_depth;
   int my_rank, size;

   (void) MPIR_Comm_size(comm, &size);
   (void) MPIR_Comm_rank(comm, &my_rank);
   my_depth = comm->Topology_Depths[my_rank];

   globus_libc_fprintf(stderr, "*** Start print comm_set from proc #%d/%d\n",
                       my_rank, size);

   for (lvl = 0; lvl < my_depth; lvl++)
   {
      comm_set_t cs = comm->Topology_CommSets[lvl];
      int i;

      globus_libc_fprintf(stderr, "lvl=%d: size=%d, root_index=%d, " \
                                  "my_rank_index=%d, set={", lvl, cs.size,
                                  cs.root_index, cs.my_rank_index);
      for (i = 0; i < cs.size; i++)
      {
         if ( i ) globus_libc_fprintf(stderr, ", ");
         globus_libc_fprintf(stderr, "%d", cs.set[i]);
      }
      globus_libc_fprintf(stderr, "}\n");
   }

   globus_libc_fprintf(stderr, "*** End print comm_set from proc #%d/%d\n",
                       my_rank, size);
   return;
}


/**********************************************************************/
/* buffer copy; copy unit is "stride" bytes */
static void
copy_buf (void * const from_buf, const int from, void * const to_buf,
          const int to, const int stride)
{
   memcpy((char *)to_buf + to*stride, (char *)from_buf + from*stride, stride);
   return;
}   /* copy_buf */


/**********************************************************************/
/* convert an absolute process number (in a communicator) into an
 * index (= position) in a cluster at the given level */
static int
proc_to_index (const int rank, int ** const Colors,
               const MPICHX_Topolevel level)
{
   int indx, i;
   int my_color = Colors[rank][level];

   for (indx = 0, i = 0; i < rank; i++)
      if ( Colors[i][level] == my_color )
         indx++;
   return indx;
}   /* proc_to_index */


/**********************************************************************/
/* A process is a master process at the given level iff its cluster.
 * ids are 0 from the given level + 1 up to its depth: a master proc
 * is the representative of a cluster at level n+1 in a cluster at
 * level n. */
static int
is_a_master (const MPICHX_Topolevel level, const int * const ClusterIds,
             const MPICHX_Topolevel depth)
{
   MPICHX_Topolevel lvl;

   for (lvl = level+1; lvl < depth; lvl++)
      if ( ClusterIds[lvl] != 0 )
         return GLOBUS_FALSE;

   return GLOBUS_TRUE;
}   /* is_a_master */


/**********************************************************************/
/* create a single set of communicating processes at a given level */
static void
make_set (const MPICHX_Topolevel lvl, struct MPIR_COMMUNICATOR * const comm)
{
   int proc, index, my_rank, size;
   int my_color;
   int ** const ClusterIds = comm->Topology_ClusterIds;
   const MPICHX_Topolevel * const Depths = comm->Topology_Depths;
   int ** const Colors = comm->Topology_Colors;
   comm_set_t * const CommSet = &(comm->Topology_CommSets[lvl]);

   (void) MPIR_Comm_size (comm, &size);
   (void) MPIR_Comm_rank (comm, &my_rank);
   my_color = Colors[my_rank][lvl];

   /* a set is made of all the master processes at the given level, ie: all
    * the root processes at level+1 */
   for (proc = 0, index = 0; proc < size; proc++)
   {
      int current_Depth = Depths[proc];
      int *current_ClusterIds = ClusterIds[proc];

      if ( lvl < current_Depth  &&  my_color == Colors[proc][lvl]  &&
           is_a_master(lvl, current_ClusterIds, current_Depth) )
      {
         /* know the index of my process in the set */
         if ( proc == my_rank )
            CommSet->my_rank_index = index;

         /* know the index of the root in this set */
         if ( current_ClusterIds[lvl] == 0 )
            CommSet->root_index = index;

         CommSet->set[index++] = proc;
      }   /* endif */
   }   /* endfor */
   CommSet->size = index;

   return;
}   /* make_set */


/**********************************************************************/
/* create the sets of processes in which I will be involved for
 * communication */
static void
update_comm_sets (struct MPIR_COMMUNICATOR * const comm)
{
   MPICHX_Topolevel first_lvl, lvl, my_depth;
   int my_rank;
   int *my_ClusterIds;

   (void) MPIR_Comm_rank (comm, &my_rank);
   my_depth = comm->Topology_Depths[my_rank];
   my_ClusterIds = comm->Topology_ClusterIds[my_rank];

   /* from which level will I be involved in a communication */
   for (first_lvl = 0, lvl = my_depth-1; lvl >= 0; lvl--)
      if ( my_ClusterIds[lvl] != 0 )
      {
         first_lvl = lvl;
         break;
      }   /* endif */

   for (lvl = 0; lvl < first_lvl; lvl++)
      (comm->Topology_CommSets[lvl]).size = 0;

   for (lvl = first_lvl; lvl < my_depth; lvl++)
      make_set(lvl, comm);

   return;
}   /* update_comm_sets */


/**********************************************************************/
/* 'rename' the clusters at each level so that the root process has
 * only zeros as cluster ids (at each level) */
static void
update_cluster_ids (const int root, struct MPIR_COMMUNICATOR * const comm)
{
   MPICHX_Topolevel lvl;
   int size;
   int *root_ClusterIds;
   const MPICHX_Topolevel * const Depths = comm->Topology_Depths;
   int ** const ClusterIds = comm->Topology_ClusterIds;
   int ** const Colors = comm->Topology_Colors;
   const int * const root_Colors = Colors[root];
   const MPICHX_Topolevel root_depth = Depths[root];

   (void) MPIR_Comm_size (comm, &size);
   root_ClusterIds = ClusterIds[root];

   for (lvl = 0; lvl < root_depth; lvl++)
   {
      int shift = root_ClusterIds[lvl];

      if ( shift )
      {
         /* at the current level, the root process has a non-zero
          * cluster id: we shift (rotate) the cids at this level for
          * all the processes which can communicate directly with the
          * root process at this level (ie: all the procs in the same
          * cluster as the root). */

         int n_cid, proc;
         const int root_color = root_Colors[lvl];

         /* find the number of cluster-ids that need to be rotated at
          * this level */
         for (n_cid = 0, proc = 0; proc < size; proc++)
            if ( Depths[proc] > lvl  &&  root_color == Colors[proc][lvl]  &&
                 ClusterIds[proc][lvl] > n_cid )
               n_cid = ClusterIds[proc][lvl];

         shift = ++n_cid - shift;

         for (proc = 0; proc < size; proc++)
            if ( Depths[proc] > lvl  &&  root_color == Colors[proc][lvl] )
               ClusterIds[proc][lvl] = (ClusterIds[proc][lvl] + shift) % n_cid;
      }   /* endif */
   }   /* endfor */

   return;
}   /* update_cluster_ids */


/**********************************************************************/
/* pack the data elements (in 'from_buf', which is a buffer allocated
 * at level 'init_lvl') process 'rank' is responsible for at level
 * 'lvl', copying them into 'to_buf' at position 'to_position'.  Also
 * update 'to_pos' for the next packs. */
static void
pack_dependencies (void * const from_buf, const int count,
                   const MPI_Datatype datatype, const MPI_Aint stride,
                   const MPICHX_Topolevel init_lvl, const int rank,
                   const MPICHX_Topolevel lvl, void * const to_buf,
                   const int to_size, int *to_position, const MPI_Comm comm,
                   const int comm_size, const MPICHX_Topolevel * const Depths,
                   int ** const Colors)
{
   const MPICHX_Topolevel next_lvl = lvl + 1;

   if ( Depths[rank] > next_lvl )
   {
      int prc;
      const int color = Colors[rank][next_lvl];

      for (prc = 0; prc < comm_size; prc++)
         /* the process 'prc' belongs to the "dependencies" of process
          * 'rank' iff they both have the same color at level 'next_lvl' */
         if ( Depths[prc] > next_lvl  &&  Colors[prc][next_lvl] == color )
            MPI_Pack(((char *)from_buf) + stride *
                                         proc_to_index(prc, Colors, init_lvl),
                     count, datatype, to_buf, to_size, to_position, comm);
   }
   else   /* process 'rank' is responsible for itself only */
      MPI_Pack(((char*)from_buf) + stride*proc_to_index(rank, Colors, init_lvl),
               count, datatype, to_buf, to_size, to_position, comm);

   return;
}   /* pack_dependencies */


/**********************************************************************/
/* unpack the data elements process 'rank' is responsible for (at
 * level 'lvl') from the buffer 'from_buf' (at position
 * 'from_position') copying them into 'to_buf' (which is a buffer
 * allocated at level 'init_lvl').  Also update 'from_position' for
 * the next data to unpack */
static void
unpack_dependencies (void * const from_buf, const int from_size,
                     int *from_position, const int rank,
                     const MPICHX_Topolevel lvl, void *to_buf, const int count,
                     const MPI_Datatype datatype, const MPI_Aint stride,
                     const MPICHX_Topolevel init_lvl, const MPI_Comm comm,
                     const int comm_size, const MPICHX_Topolevel * const Depths,
                     int ** const Colors)
{
   MPICHX_Topolevel next_lvl = lvl + 1;

   if ( Depths[rank] > next_lvl )
   {
      int prc;
      const int color = Colors[rank][next_lvl];

      for (prc = 0; prc < comm_size; prc++)
         /* the process 'prc' belongs to the "dependencies" of process
          * 'rank' iff they both have the same color at level 'next_lvl' */
         if ( Depths[prc] > next_lvl  &&  Colors[prc][next_lvl] == color )
            MPI_Unpack(from_buf, from_size, from_position,
                       ((char *)to_buf) + stride *
                                         proc_to_index(prc, Colors, init_lvl),
                       count, datatype, comm);
   }
   else
      MPI_Unpack(from_buf, from_size, from_position,
                 ((char*)to_buf) + stride*proc_to_index(rank, Colors, init_lvl),
                 count, datatype, comm);

   return;
}   /* unpack_dependencies */


/**********************************************************************/
/* perform an MPI_Waitall on the given number of MPI_Request's */
static int
wait_for_all_reqs (MPI_Request * const req, const int n_req)
{
   MPI_Status *statuses;
   int mpi_errno = MPI_SUCCESS;

   if ( !n_req ) return MPI_SUCCESS;

   g_malloc_chk(statuses, MPI_Status *, sizeof(MPI_Status) * n_req);
   mpi_errno = MPI_Waitall(n_req, req, statuses);
   /* set mpi_errno in function of statuses if mpi_errno != MPI_SUCCESS */

   g_free(statuses);

   return mpi_errno;
}   /* wait_for_all_reqs */


/**********************************************************************/
#ifdef MPID_Barrier
static int
flat_tree_enter_barrier (const comm_set_t comm_set, const MPI_Comm comm)
{
   int mpi_errno = MPI_SUCCESS;
   const int set_size = comm_set.size;
   const int my_rank_idx = comm_set.my_rank_index;
   const int root_idx = comm_set.root_index;
   const int * const set = comm_set.set;

   /* any fellow to talk to? */
   if ( set_size < 2 ) return MPI_SUCCESS;

   if ( my_rank_idx == root_idx )   /* I'm the root of the set */
   {
      int i;
      MPI_Request *req;
      int n_req = 0;

      g_malloc_chk(req, MPI_Request *, sizeof(MPI_Request) * (set_size-1));
      /* wait for all the notifications from the procs */
      for (i = 0; i < set_size; i++)
         if ( i != my_rank_idx )
         {
            mpi_errno = MPI_Irecv((void *)0, 0, MPI_INT, set[i],
                                  MPIR_BARRIER_TAG, comm, req + n_req++);
            if ( mpi_errno ) return mpi_errno;
         }   /* endif */
      mpi_errno = wait_for_all_reqs(req, n_req);
      g_free(req);
   }
   else   /* I'm not the root of this set */
   {
      /* notify the root I reached the barrier */
      mpi_errno = MPI_Send((void *)0, 0, MPI_INT, set[root_idx],
                           MPIR_BARRIER_TAG, comm);
   }   /* endif */

   return mpi_errno;
}   /* flat_tree_enter_barrier */
#endif


/**********************************************************************/
#ifdef MPID_Barrier
static int
flat_tree_exit_barrier (const comm_set_t comm_set, const MPI_Comm comm)
{
   int mpi_errno = MPI_SUCCESS;
   const int my_rank_idx = comm_set.my_rank_index;
   const int root_idx = comm_set.root_index;
   const int set_size = comm_set.size;
   const int * const set = comm_set.set;

   /* any fellow to talk to? */
   if ( set_size < 2 ) return MPI_SUCCESS;

   if ( my_rank_idx == root_idx )   /* I'm the root of the set */
   {
      int i;
      MPI_Request *req;
      int n_req = 0;

      g_malloc_chk(req, MPI_Request *, sizeof(MPI_Request) * (set_size-1));
      /* send a GO signal the all the processes in my set */
      for (i = 0; i < set_size; i++)
         if ( i != my_rank_idx )
         {
            mpi_errno = MPI_Isend((void *)0, 0, MPI_INT, set[i],
                                  MPIR_BARRIER_TAG, comm, req + n_req++);
            if ( mpi_errno ) return mpi_errno;
         }   /* endif */
      mpi_errno = wait_for_all_reqs(req, n_req);
      g_free(req);
   }
   else   /* I'm not the root of this set */
   {
      MPI_Status status;

      /* wait for root's GO signal */
      mpi_errno = MPI_Recv((void *)0, 0, MPI_INT, set[root_idx],
                           MPIR_BARRIER_TAG, comm, &status);
   }   /* endif */

   return mpi_errno;
}   /* flat_tree_exit_barrier */
#endif


/**********************************************************************/
#ifdef MPID_Bcast
static int
binomial_bcast (void *buffer, const int count, const MPI_Datatype datatype,
                const MPI_Comm comm, const comm_set_t set)
{
   int mask, relative_rnk_idx;
   int mpi_errno = MPI_SUCCESS;
   MPI_Request *req;
   int n_req = 0;
   const int my_rank_idx = set.my_rank_index;
   const int root_idx = set.root_index;
   const int set_size = set.size;

   /* any fellow to talk to? */
   if ( set_size < 2 ) return MPI_SUCCESS;

   relative_rnk_idx = (my_rank_idx - root_idx + set_size) % set_size;

   mask = 0x1;
   while ( mask < set_size )
   {
      if ( relative_rnk_idx & mask )
      {
         MPI_Status st;
         const int src_index = (my_rank_idx - mask + set_size) % set_size;

         mpi_errno = MPI_Recv(buffer, count, datatype, set.set[src_index],
                              MPIR_BCAST_TAG, comm, &st);
         if (mpi_errno) return mpi_errno;
         break;
      }   /* endif */
      mask <<= 1;
   }   /* end while */

   g_malloc_chk(req, MPI_Request *, sizeof(MPI_Request) * set_size);

   /* using the binomial tree algorithm, I may have to relay the message */
   mask >>= 1;
   while (mask > 0)
   {
      if (relative_rnk_idx + mask < set_size)
      {
         const int dst_index = (my_rank_idx + mask) % set_size;

         mpi_errno = MPI_Isend(buffer, count, datatype, set.set[dst_index],
                               MPIR_BCAST_TAG, comm, req + n_req++);
         if ( mpi_errno ) return mpi_errno;
      }
      mask >>= 1;
   }   /* end while */

   mpi_errno = wait_for_all_reqs(req, n_req);
   g_free(req);

   return mpi_errno;
}   /* binomial_bcast */
#endif


/**********************************************************************/
/* flat tree BroadCast: the root process sends the msg to the each of
 * the processes in its set */
#ifdef MPID_Bcast
static int
flat_tree_bcast (void *buffer, const int count, const MPI_Datatype datatype,
                 const MPI_Comm comm, const comm_set_t set)
{
   int mpi_errno = MPI_SUCCESS;
   const int root_idx = set.root_index;
   const int my_rank_idx = set.my_rank_index;
   const int set_size = set.size;

   /* any fellow to talk to? */
   if ( set_size < 2 ) return MPI_SUCCESS;

   if ( root_idx == my_rank_idx )   /* I'm root, i send */
   {
      int i;

      for (i = 0; i < set_size; i++)
         if ( i != my_rank_idx )
         {
            mpi_errno = MPI_Send(buffer, count, datatype, set.set[i],
                                 MPIR_BCAST_TAG, comm);
            if (mpi_errno) return mpi_errno;
         }   /* endif */
   }
   else   /* I'm not the root proc in this set, I recv */
   {
      MPI_Status status;

      mpi_errno = MPI_Recv(buffer, count, datatype, set.set[root_idx],
                           MPIR_BCAST_TAG, comm, &status);
   }   /* endif */

   return mpi_errno;
}   /* flat_tree_bcast */
#endif


/**********************************************************************/
#ifdef MPID_Gather
static int
flat_tree_gather (void * const my_buf, const int sendcnt,
                  const MPI_Datatype sendtype, const int recvcnt,
                  const MPI_Datatype recvtype, const MPI_Comm comm,
                  const int comm_size, void *tmp_buf, const int buf_size,
                  const MPI_Aint stride, const MPICHX_Topolevel init_lvl,
                  const comm_set_t set, const MPICHX_Topolevel * const Depths,
                  int ** const Colors, const MPICHX_Topolevel lvl)
{
   int mpi_errno = MPI_SUCCESS;
   const int my_rank_idx = set.my_rank_index;
   const int root_idx = set.root_index;
   const int set_size = set.size;

   /* any fellow to talk to? */
   if ( set_size < 2 ) return MPI_SUCCESS;

   if ( root_idx == my_rank_idx )   /* I'm the root of this set: I recv */
   {
      int i;

      for (i = 0; i < set_size; i++)   /* I recv lots */
         if ( i != root_idx )
         {
            const int src = set.set[i];
            int pos = 0;
            MPI_Status status;

            mpi_errno = MPI_Recv(tmp_buf, buf_size, MPI_PACKED, src,
                                 MPIR_GATHER_TAG, comm, &status);
            if ( mpi_errno ) return mpi_errno;

            /* copy the elements recv'd in tmp_buf to their correct position
             * in my_buf */
            unpack_dependencies(tmp_buf, buf_size, &pos, src, lvl, my_buf,
                                recvcnt, recvtype, stride, init_lvl, comm,
                                comm_size, Depths, Colors);
         }   /* endif */
   }
   else   /* I'm not the root process: I send my buffer to the local root */
   {
      int pos = 0;

      pack_dependencies(my_buf, sendcnt, sendtype, stride, init_lvl,
                        set.set[my_rank_idx], lvl, tmp_buf, buf_size, &pos,
                        comm, comm_size, Depths, Colors);
      mpi_errno = MPI_Send(tmp_buf, pos, MPI_PACKED, set.set[root_idx],
                           MPIR_GATHER_TAG, comm);
   }

   return mpi_errno;
}   /* flat_tree_gather */
#endif


/**********************************************************************/
#ifdef MPID_Gather
static int
binomial_gather (void * const my_buf, const int sendcnt,
                 const MPI_Datatype sendtype, const int recvcnt,
                 const MPI_Datatype recvtype, const MPI_Comm comm,
                 const int comm_size, void *tmp_buf, const int buf_size,
                 const MPI_Aint stride, const MPICHX_Topolevel init_lvl,
                 const comm_set_t set, const MPICHX_Topolevel * const Depths,
                 int ** const Colors, const MPICHX_Topolevel lvl)
{
   int mpi_errno = MPI_SUCCESS;
   int relative_rnk_idx, mask;
   const int my_rank_idx = set.my_rank_index;
   const int root_idx = set.root_index;
   const int set_size = set.size;

   /* any fellow to talk to? */
   if ( set_size < 2 ) return MPI_SUCCESS;

   relative_rnk_idx = (my_rank_idx - root_idx + set_size) % set_size;
   mask = 0x1;
   /* receive some chunks of data and copy them into my own buffer */
   while ( mask < set_size )
   {
      const int rel_src_idx = relative_rnk_idx + mask;

      if ( relative_rnk_idx & mask ) break;

      if ( rel_src_idx < set_size )
      {
         MPI_Status status;
         int i, pos;
         const int source = set.set[(rel_src_idx + root_idx) % set_size];

         mpi_errno = MPI_Recv(tmp_buf, buf_size, MPI_PACKED, source,
                              MPIR_GATHER_TAG, comm, &status);
         if ( mpi_errno ) return mpi_errno;
         /* copy the recv'd elements from tmp_buf to my_buf */
         pos = 0;
         for (i = 0; i < mask; i++)
         {
            int src;

            src = rel_src_idx + i;
            if ( src >= set_size ) break;
            src = set.set[(src + root_idx) % set_size];
            unpack_dependencies(tmp_buf, buf_size, &pos, src, lvl, my_buf,
                                recvcnt, recvtype, stride, init_lvl, comm,
                                comm_size, Depths, Colors);
         }   /* endfor */
      }   /* endif */
      mask <<= 1;
   }   /* end while */

   /* send all the data elements I collected to my local root */
   if ( my_rank_idx != root_idx )
   {
      int i, pos = 0;
      const int dst = (my_rank_idx - mask + set_size) % set_size;

      for (i = 0; i < mask; i++)
      {
         int prc = relative_rnk_idx + i;

         if ( prc >= set_size ) break;
         prc = set.set[(prc + root_idx) % set_size];
         pack_dependencies(my_buf, sendcnt, sendtype, stride, init_lvl, prc,
                           lvl, tmp_buf, buf_size, &pos, comm, comm_size,
                           Depths, Colors);
      }   /* endfor */
      mpi_errno = MPI_Send(tmp_buf, pos, MPI_PACKED, set.set[dst],
                           MPIR_GATHER_TAG, comm);
   }   /* endif */

   return mpi_errno;
}   /* binomial_gather */
#endif


/**********************************************************************/
#ifdef MPID_Scatter
static int
flat_tree_scatter (void * const my_buf, const int sendcnt,
                   const MPI_Datatype sendtype, const int recvcnt,
                   const MPI_Datatype recvtype, const MPI_Comm comm,
                   const int comm_size, void *tmp_buf, const int buf_size,
                   const MPI_Aint stride, const MPICHX_Topolevel init_lvl,
                   const comm_set_t set, const MPICHX_Topolevel * const Depths,
                   int ** const Colors, const MPICHX_Topolevel lvl)
{
   int mpi_errno = MPI_SUCCESS;
   const int my_rank_idx = set.my_rank_index;
   const int root_idx = set.root_index;
   const int set_size = set.size;

   /* any fellow to talk to? */
   if ( set_size < 2 ) return MPI_SUCCESS;

   if ( root_idx == my_rank_idx )   /* send chunks to the procs of this set */
   {
      int i;

      for (i = 0; i < set_size; i++)
         if ( i != my_rank_idx )
         {
            const int dst = set.set[i];
            int pos = 0;

            /* copy the right chunks of my_buf into tmp_buf */
            pack_dependencies(my_buf, sendcnt, sendtype, stride, init_lvl, dst,
                              lvl, tmp_buf, buf_size, &pos, comm, comm_size,
                              Depths, Colors);

            /* send my data */
            mpi_errno = MPI_Send(tmp_buf, pos, MPI_PACKED, dst,
                                 MPIR_SCATTER_TAG, comm);
            if ( mpi_errno ) return mpi_errno;
         }   /* endif */
   }
   else   /* I recv from the root of this set of processes */
   {
      MPI_Status status;
      int pos = 0;

      mpi_errno = MPI_Recv(tmp_buf, buf_size, MPI_PACKED, set.set[root_idx],
                           MPIR_SCATTER_TAG, comm, &status);
      if ( mpi_errno ) return mpi_errno;
      unpack_dependencies(tmp_buf, buf_size, &pos, set.set[my_rank_idx], lvl,
                          my_buf, recvcnt, recvtype, stride, init_lvl, comm,
                          comm_size, Depths, Colors);
   }   /* endif */

   return mpi_errno;
}   /* flat_tree_scatter */
#endif


/**********************************************************************/
#ifdef MPID_Scatter
static int
binomial_scatter (void * const my_buf, const int sendcnt,
                  const MPI_Datatype sendtype, const int recvcnt,
                  const MPI_Datatype recvtype, const MPI_Comm comm,
                  const int comm_size, void *tmp_buf, const int buf_size,
                  const MPI_Aint stride, const MPICHX_Topolevel init_lvl,
                  const comm_set_t set, const MPICHX_Topolevel * const Depths,
                  int ** const Colors, const MPICHX_Topolevel lvl)
{
   int mpi_errno = MPI_SUCCESS;
   int mask, relative_rnk_idx;
   const int my_rank_idx = set.my_rank_index;
   const int root_idx = set.root_index;
   const int set_size = set.size;

   /* any fellow to talk to? */
   if ( set_size < 2 ) return MPI_SUCCESS;

   relative_rnk_idx = (my_rank_idx - root_idx + set_size) % set_size;
   mask = 0x1;

   /* find the guy which is going to send me the data */
   while ( mask < set_size )
   {
      if ( relative_rnk_idx & mask ) break;
      mask <<= 1;
   }

   /* receive some chunks of data from my local root */
   if ( my_rank_idx != root_idx )
   {
      MPI_Status status;
      int i;
      const int source = set.set[(my_rank_idx - mask + set_size) % set_size];
      int pos = 0;

      mpi_errno = MPI_Recv(tmp_buf, buf_size, MPI_PACKED, source,
                           MPIR_SCATTER_TAG, comm, &status);
      if ( mpi_errno ) return mpi_errno;
      for (i = 0; i < mask; i++)
      {
         int prc = relative_rnk_idx + i;

         if ( prc >= set_size ) break;
         prc = set.set[(prc + root_idx) % set_size];
         unpack_dependencies(tmp_buf, buf_size, &pos, prc, lvl, my_buf,
                             recvcnt, recvtype, stride, init_lvl, comm,
                             comm_size, Depths, Colors);
      }
   }   /* endif */

   mask >>= 1;
   /* dispatch data elements to my fellows (binomial-tree algorithm) */
   while ( mask > 0 )
   {
      int dst = relative_rnk_idx + mask;

      if ( dst < set_size )
      {
          int i, pos = 0;

         /* copy the right data elements from my_buf into tmp_buf */
         for (i = 0; i < mask; i++)
         {
            int sub_dest = dst + i;

            if ( sub_dest >= set_size ) break;
            sub_dest = set.set[(sub_dest + root_idx) % set_size];
            pack_dependencies(my_buf, sendcnt, sendtype, stride, init_lvl,
                              sub_dest, lvl, tmp_buf, buf_size, &pos, comm,
                              comm_size, Depths, Colors);
         }   /* endfor */

         dst = set.set[(dst + root_idx) % set_size];
         /* send the tmp_buf */
         mpi_errno = MPI_Send(tmp_buf, pos, MPI_PACKED, dst, MPIR_SCATTER_TAG,
                              comm);
         if ( mpi_errno ) return mpi_errno;
      }
      mask >>= 1;
   }   /* end while */

   return mpi_errno;
}   /* binomial_scatter */
#endif


/**********************************************************************/
#ifdef MPID_Reduce
static int
flat_tree_reduce (comm_set_t set, void *my_buf, int extent, int lb, int count,
                  MPI_Datatype datatype, MPI_Comm comm, MPI_User_function *uop,
                  void *tmp_buf)
{
   /* here the operation is assumed to be commutative and associative */

   int mpi_errno = MPI_SUCCESS;
   int root_idx, rank_idx;

   root_idx = set.root_index;
   rank_idx = set.my_rank_index;

   /* I recv and compute only if I'm the root of this set */
   if ( root_idx == rank_idx )
   {
      int i, set_size;

      set_size = set.size;
      if ( set_size < 2 ) return MPI_SUCCESS;

      /* receive all the elements and compute */
      for (i = 0; i < set_size; i++)
         if ( i != root_idx )
         {
            MPI_Status status;

            mpi_errno = MPI_Recv((char *) tmp_buf, count, datatype, set.set[i],
                                 MPIR_REDUCE_TAG, comm, &status);
            if ( mpi_errno ) return mpi_errno;
            /* compute */
            (*uop)(tmp_buf, my_buf, &count, &datatype);
         }   /* endif */
   }
   else   /* I'm not root: I send my buffer */
      mpi_errno = MPI_Send(my_buf, count, datatype, set.set[root_idx],
                           MPIR_REDUCE_TAG, comm);

   return mpi_errno;
}   /* flat_tree_reduce */
#endif


/**********************************************************************/
#ifdef MPID_Reduce
static int
hypercube_reduce (comm_set_t set, void *my_buf, int count,
                  MPI_Datatype datatype, MPI_Comm comm, int extent, int lb,
                  MPI_User_function *uop, void *tmp_buf)
{
   /* here the operation is assumed to be commutative and associative */

   int mpi_errno = MPI_SUCCESS;
   int relative_rnk_idx, mask, root_idx, rank_idx, set_size;

   root_idx = set.root_index;
   rank_idx = set.my_rank_index;
   set_size = set.size;

   relative_rnk_idx = (rank_idx - root_idx + set_size) % set_size;
   mask = 0x1;
   while ( mask < set_size )
   {
      if ( mask & relative_rnk_idx )   /* send my (intermediate) result */
      {
         int dst = set.set[(rank_idx - mask + set_size) % set_size];

         mpi_errno = MPI_Send(my_buf, count, datatype, dst, MPIR_REDUCE_TAG,
                              comm);
         break;
      }
      else   /* recv and compute */
      {
         int source;

         source = relative_rnk_idx | mask;
         if ( source < set_size )
         {
            MPI_Status status;

            source = set.set[(source + root_idx) % set_size];
            mpi_errno = MPI_Recv(tmp_buf, count, datatype, source,
                                 MPIR_REDUCE_TAG, comm, &status);
            if ( mpi_errno ) break;
            /* compute */
            (*uop)(tmp_buf, my_buf, &count, &datatype);
         }   /* endif */
      }
      mask <<= 1;
   }   /* end while */

   return mpi_errno;
}   /* hypercube_reduce */
#endif



/**********************************************************************/
/* PUBLIC FUNCTIONS                                                   */
/**********************************************************************/


/**********************************************************************/
#ifdef MPID_Barrier
int
MPID_FN_Barrier (struct MPIR_COMMUNICATOR *comm)
{
   int mpi_errno = MPI_SUCCESS;
   MPICHX_Topolevel lvl;
   int size, my_rank, my_depth;
   comm_set_t *CommSets;

   /* Intialize communicator size */
   (void) MPIR_Comm_size (comm, &size);

   /* If there's only one member, this is trivial */
   if ( size == 1 ) return MPI_SUCCESS;

   /* Switch communicators to the hidden collective */
   comm = comm->comm_coll;

   (void) MPIR_Comm_rank(comm, &my_rank);
   my_depth = comm->Topology_Depths[my_rank];

   update_comm_sets(comm);
   CommSets = comm->Topology_CommSets;

   /* Lock for collective operation */
   MPID_THREAD_LOCK(comm->ADIctx,comm);

   /* enter the barrier (tell the people I reached the barrier) */
   for (lvl = my_depth-1; lvl >= 0; lvl--)
   {
      mpi_errno = flat_tree_enter_barrier(CommSets[lvl], comm->self);
      if ( mpi_errno ) goto clean_exit;
   }

   /* exit the barrier (tell the people they can go on working) */
   for (lvl = 0; lvl < my_depth; lvl++)
   {
      mpi_errno = flat_tree_exit_barrier(CommSets[lvl], comm->self);
      if ( mpi_errno ) goto clean_exit;
   }   /* endfor */

clean_exit:
   /* Unlock for collective operation */
   MPID_THREAD_UNLOCK(comm->ADIctx,comm);

   return mpi_errno;
}   /* MPID_FN_Barrier */
#endif   /* MPID_Barrier */


/**********************************************************************/
#ifdef MPID_Bcast
int
MPID_FN_Bcast (void *buffer, int count, struct MPIR_DATATYPE *datatype,
               int root, struct MPIR_COMMUNICATOR *comm)
{
   MPICHX_Topolevel lvl, my_depth;
   int size, my_rank;
   int mpi_errno = MPI_SUCCESS;
   static char myname[] = "MPI_BCAST";
   comm_set_t *CommSets;

   /* Is root within the comm and more than 1 processes involved? */
   (void) MPIR_Comm_size (comm, &size);

#ifndef MPIR_NO_ERROR_CHECKING
   if (root >= size) 
      mpi_errno = MPIR_Err_setmsg (MPI_ERR_ROOT, MPIR_ERR_ROOT_TOOBIG, 
                                   myname, (char *)0, (char *)0, root, size);
   else if (root < 0)
      mpi_errno = MPIR_Err_setmsg (MPI_ERR_ROOT, MPIR_ERR_DEFAULT, myname,
                                   (char *)0, (char*)0, root);
   if (mpi_errno)
      return MPIR_ERROR(comm, mpi_errno, myname);
#endif

   /* If there is only one process or nothing to broadcast... */
   if ( size == 1  ||  count == 0 ) return MPI_SUCCESS;

   /* Switch communicators to the hidden collective */
   comm = comm->comm_coll;

   /* Algorithm:
    * using the cluster IDs, we guess to which processes we need to
    * send messages, or from which proc we need to recv a msg.  We
    * start communicating thru the slowest links (WAN-TCP), ending up
    * with the fastest communication level (vMPI or localhost-TCP).
    * At each communication level, a local root process broadcasts the
    * msg to the representatives of the other cluster at the current
    * level, using either a flat tree algo or a binomial tree algo.
    * For high-latency networks (WAN TCP), a flat tree algo is better.
    * Otherwise, we use a binomial tree algorithm. */

   /* first we 'rename' the clusters at each level so that the root
    * process have only zeros as cluster IDs (at each level). */
   update_cluster_ids(root, comm);
   update_comm_sets(comm);

   (void) MPIR_Comm_rank (comm, &my_rank);
   my_depth = comm->Topology_Depths[my_rank];
   CommSets = comm->Topology_CommSets;

   /* Lock for collective operation */
   MPID_THREAD_LOCK (comm->ADIctx,comm);

   for (lvl = 0; lvl < my_depth; lvl++)
   {
      if ( lvl == MPICHX_WAN_LEVEL )   /* flat-tree algorithm */
         mpi_errno = flat_tree_bcast(buffer, count, datatype->self, comm->self,
                                     CommSets[lvl]);
      else   /* binomial tree algorithm */
         mpi_errno = binomial_bcast(buffer, count, datatype->self, comm->self,
                                    CommSets[lvl]);
      if (mpi_errno) break;
   }   /* endfor */

   /* Unlock for collective operation */
   MPID_THREAD_UNLOCK(comm->ADIctx,comm);

   return mpi_errno;
}   /* MPID_FN_Bcast */
#endif   /* MPID_Bcast */


/**********************************************************************/
#ifdef MPID_Gather
int
MPID_FN_Gather (void *sendbuf, int sendcnt, struct MPIR_DATATYPE *sendtype,
                void *recvbuf, int recvcnt, struct MPIR_DATATYPE *recvtype,
                int root, struct MPIR_COMMUNICATOR *comm)
{
   int mpi_errno = MPI_SUCCESS;
   void *tmp_buf;
   int size, my_rank;
   MPICHX_Topolevel lvl, init_lvl, my_depth;
   const MPICHX_Topolevel * const Depths = comm->Topology_Depths;
   static char myname[] = "MPI_GATHER";
   comm_set_t *CommSets;
   int ** const Colors = comm->Topology_Colors;
   MPI_Status status;
   MPI_Aint buf_size, stride, sendtype_extent, lb, ub;

   /* Is root within the communicator? */
   (void) MPIR_Comm_size (comm, &size);

#ifndef MPIR_NO_ERROR_CHECKING
   if ( root >= size )
      mpi_errno = MPIR_Err_setmsg (MPI_ERR_ROOT, MPIR_ERR_ROOT_TOOBIG, myname,
                                   (char *)0, (char *)0, root, size);
   else if ( root < 0 )
      mpi_errno = MPIR_Err_setmsg (MPI_ERR_ROOT, MPIR_ERR_DEFAULT, myname,
                                   (char *)0, (char *)0, root);
   if ( mpi_errno )
      return MPIR_ERROR(comm, mpi_errno, myname);
#endif

   /* Get my rank and switch communicators to the hidden collective */
   (void) MPIR_Comm_rank (comm, &my_rank);
   comm = comm->comm_coll;

   /* first we 'rename' the clusters at each level so that the root
    * process has only zeros as cluster IDs (at each level). */
   update_cluster_ids(root, comm);
   update_comm_sets(comm);
   CommSets = comm->Topology_CommSets;
   my_depth = Depths[my_rank];

   /* Find the 1st level (init_lvl) of the communication sets I will
    * be involved in */
   for (init_lvl = 0; init_lvl < my_depth; init_lvl++)
      if ( CommSets[init_lvl].size != 0 ) break;

   MPIR_Type_get_limits(sendtype, &lb, &ub);
   sendtype_extent = ub - lb;
   stride = sendcnt * sendtype_extent;
   buf_size = comm->Topology_ClusterSizes[init_lvl] * stride;

   /* MPI standard: recvbuf, recvcnt, recvtype are not significant for
    * procs != root.  If I'm not root, allocate memory to hold data I
    * may have to relay from processes I'm responsible for */
   if ( my_rank != root )
   {
      recvcnt = sendcnt;
      recvtype = sendtype;
      g_malloc_chk(recvbuf, void *, buf_size);
      recvbuf = (void*) ((char*) recvbuf - lb);
   }

   /* Lock for collective operation */
   MPID_THREAD_LOCK(comm->ADIctx,comm);

   /* put the data I currently have to relay into recvbuf */
   /* MPI_Sendrecv should be optimized in case source == destination */
   mpi_errno = MPI_Sendrecv(sendbuf, sendcnt, sendtype->self, my_rank,
                            MPIR_GATHER_TAG, ((char*)recvbuf) + stride *
                                     proc_to_index(my_rank, Colors, init_lvl),
                            recvcnt, recvtype->self, my_rank, MPIR_GATHER_TAG,
                            comm->self, &status);
   if ( mpi_errno ) goto clean_exit;

   /* allocate a temporary buffer to recv packed data and where to
    * pack data before sending */
   g_malloc_chk(tmp_buf, void *, buf_size);

   /* p: # of procs; o: overhead to send/recv; l: latency
    * flat tree: time = o*p + l
    * binomial tree: time = (l+2*o) * ceil(log p)
    * If l>>o: for large values of p, binomial tree is more efficient;
    *          for small values of p, flat tree is faster;
    * If l~=o: idem, but the threshold for 'p' is lower. */

   /* depending on the msg size (sendcnt * extent * size), it might be
    * more efficient:
    *  - to take advantage of the protocol levels in case of small msg
    *    size (because there are memory copies),
    *  - or to ignore the protocol levels in case of large msg size
    *    (when the memory copies would take more time than latencies). */

   for (lvl = my_depth-1; lvl >= 0; lvl--)
   {
      if ( lvl == MPICHX_WAN_LEVEL )   /* WAN TCP: flat tree */
         mpi_errno = flat_tree_gather(recvbuf, sendcnt, sendtype->self,
                                      recvcnt, recvtype->self, comm->self,
                                      size, tmp_buf, buf_size, stride, init_lvl,
                                      CommSets[lvl], Depths, Colors, lvl);
      else   /* binomial tree algorithm */
         mpi_errno = binomial_gather(recvbuf, sendcnt, sendtype->self, recvcnt,
                                     recvtype->self, comm->self, size, tmp_buf,
                                     buf_size, stride, init_lvl, CommSets[lvl],
                                     Depths, Colors, lvl);
      if ( mpi_errno ) break;
   }   /* endfor */

   /* Unlock for collective operation */
   MPID_THREAD_UNLOCK(comm->ADIctx,comm);

   g_free(tmp_buf);

clean_exit:
   if ( my_rank != root )
      g_free(recvbuf);

   return mpi_errno;
}   /* MPID_FN_Gather */
#endif   /* MPID_Gather */


/**********************************************************************/
#ifdef MPID_Gatherv
int
MPID_FN_Gatherv (void *sendbuf, int sendcnt, struct MPIR_DATATYPE *sendtype,
                 void *recvbuf, int *recvcnts, int *displs,
                 struct MPIR_DATATYPE *recvtype, int root,
                 struct MPIR_COMMUNICATOR *comm)
{
   int mpi_errno = MPI_SUCCESS;
   int size, rank;
   static char myname[] = "MPI_GATHERV";

   /* Is root within the communicator? */
   (void) MPIR_Comm_size (comm, &size);

#ifndef MPIR_NO_ERROR_CHECKING
   if ( root >= size )
      mpi_errno = MPIR_Err_setmsg (MPI_ERR_ROOT, MPIR_ERR_ROOT_TOOBIG, myname,
                                   (char *)0, (char *)0, root, size);
   else if ( root < 0 )
      mpi_errno = MPIR_Err_setmsg (MPI_ERR_ROOT, MPIR_ERR_DEFAULT, myname,
                                   (char *)0, (char *)0, root);
   if ( mpi_errno )
      return MPIR_ERROR(comm, mpi_errno, myname);
#endif

   /* Get my rank and switch communicators to the hidden collective */
   (void) MPIR_Comm_rank (comm, &rank);
   comm = comm->comm_coll;

   /* Lock for collective operation */
   MPID_THREAD_LOCK(comm->ADIctx,comm);

/* do something here! */

   /* Unlock for collective operation */
   MPID_THREAD_UNLOCK(comm->ADIctx,comm);

   return mpi_errno;
}   /* MPID_FN_Gatherv */
#endif   /* MPID_Gatherv */


/**********************************************************************/
#ifdef MPID_Scatter
int
MPID_FN_Scatter (void *sendbuf, int sendcnt, struct MPIR_DATATYPE *sendtype,
                 void *recvbuf, int recvcnt, struct MPIR_DATATYPE *recvtype,
                 int root, struct MPIR_COMMUNICATOR *comm)
{
   int mpi_errno = MPI_SUCCESS;
   void *tmp_buf;
   int size, my_rank;
   MPICHX_Topolevel init_lvl, lvl, my_depth;
   const MPICHX_Topolevel * const Depths = comm->Topology_Depths;
   static char myname[] = "MPI_SCATTER";
   comm_set_t *CommSets;
   int ** const Colors = comm->Topology_Colors;
   MPI_Status status;
   MPI_Aint buf_size, stride, recvtype_extent, ub, lb;

   /* Is root within the communicator? */
   (void) MPIR_Comm_size (comm, &size);

#ifndef MPIR_NO_ERROR_CHECKING
   if ( root >= size )
      mpi_errno = MPIR_Err_setmsg (MPI_ERR_ROOT, MPIR_ERR_ROOT_TOOBIG, myname,
                                   (char *)0, (char *)0, root, size);
   else if ( root < 0 )
      mpi_errno = MPIR_Err_setmsg (MPI_ERR_ROOT, MPIR_ERR_DEFAULT, myname,
                                   (char *)0, (char *)0, root);
   if ( mpi_errno )
      return MPIR_ERROR(comm, mpi_errno, myname);
#endif

   /* Get my rank and switch communicators to the hidden collective */
   (void) MPIR_Comm_rank (comm, &my_rank);
   comm = comm->comm_coll;

   /* first we 'rename' the clusters at each level so that the root
    * process has only zeros as cluster IDs (at each level). */
   update_cluster_ids(root, comm);
   update_comm_sets(comm);
   CommSets = comm->Topology_CommSets;
   my_depth = Depths[my_rank];

   /* Find the 1st level (init_lvl) of the communication sets I will
    * be involved in */
   for (init_lvl = 0; init_lvl < my_depth; init_lvl++)
      if ( CommSets[init_lvl].size != 0 ) break;

   MPIR_Type_get_limits(recvtype, &lb, &ub);
   recvtype_extent = ub - lb;
   stride = recvcnt * recvtype_extent;
   buf_size = comm->Topology_ClusterSizes[init_lvl] * stride;

   /* MPI Standard: 'sendbuf', 'sendcnt', 'sendtype' are significant only
    * at root.  If I'm not root, allocate memory to hold data I may
    * have to relay from root to the processes I'm responsible for */
   if ( my_rank != root )
   {
      sendtype = recvtype;
      sendcnt = recvcnt;
      g_malloc_chk(sendbuf, void *, buf_size);
      sendbuf = (void*) ((char*) sendbuf - lb);
   }

   /* allocate a temporary buffer to recv packed data and where to
    * pack data before sending */
   g_malloc_chk(tmp_buf, void *, buf_size);

   /* Lock for collective operation */
   MPID_THREAD_LOCK(comm->ADIctx,comm);

   for (lvl = 0; lvl < my_depth; lvl++)
   {
      if ( lvl == MPICHX_WAN_LEVEL )   /* flat-tree algorithm */
         mpi_errno = flat_tree_scatter(sendbuf, sendcnt, sendtype->self,
                                       recvcnt, recvtype->self, comm->self,
                                       size, tmp_buf, buf_size, stride,
                                       init_lvl, CommSets[lvl], Depths, Colors,
                                       lvl);
      else   /* low latency ==> binomial-tree algorithm */
         mpi_errno = binomial_scatter(sendbuf, sendcnt, sendtype->self, recvcnt,
                                      recvtype->self, comm->self, size, tmp_buf,
                                      buf_size, stride, init_lvl, CommSets[lvl],
                                      Depths, Colors, lvl);
      if ( mpi_errno ) break;
   }   /* endfor */

   /* MPI_Sendrecv should be optimized in case source == destination */
   mpi_errno = MPI_Sendrecv(((char *) sendbuf) + stride *
                                     proc_to_index(my_rank, Colors, init_lvl),
                            sendcnt, sendtype->self, my_rank, MPIR_SCATTER_TAG,
                            recvbuf, recvcnt, recvtype->self, my_rank,
                            MPIR_SCATTER_TAG, comm->self, &status);

   /* Unlock for collective operation */
   MPID_THREAD_UNLOCK(comm->ADIctx,comm);

   if ( my_rank != root )
      g_free(sendbuf);
   g_free(tmp_buf);

   return mpi_errno;
}   /* MPID_FN_Scatter */
#endif   /* MPID_Scatter */


/**********************************************************************/
#ifdef MPID_Scatterv
int
MPID_FN_Scatterv (void *sendbuf, int *sendcnts, int *displs,
                  struct MPIR_DATATYPE *sendtype, void *recvbuf,
                  int recvcnt, struct MPIR_DATATYPE *recvtype, int root,
                  struct MPIR_COMMUNICATOR *comm)
{
   int mpi_errno = MPI_SUCCESS;
   int size, rank;
   static char myname[] = "MPI_SCATTERV";

   /* Is root within the communicator? */
   (void) MPIR_Comm_size (comm, &size);

#ifndef MPIR_NO_ERROR_CHECKING
   if ( root >= size )
      mpi_errno = MPIR_Err_setmsg (MPI_ERR_ROOT, MPIR_ERR_ROOT_TOOBIG, myname,
                                   (char *)0, (char *)0, root, size);
   else if ( root < 0 )
      mpi_errno = MPIR_Err_setmsg (MPI_ERR_ROOT, MPIR_ERR_DEFAULT, myname,
                                   (char *)0, (char *)0, root);
   if ( mpi_errno )
      return MPIR_ERROR(comm, mpi_errno, myname);
#endif

   /* Get my rank and switch communicators to the hidden collective */
   (void) MPIR_Comm_rank (comm, &rank);
   comm = comm->comm_coll;

   /* Lock for collective operation */
   MPID_THREAD_LOCK(comm->ADIctx,comm);

/* do something here! */

   /* Unlock for collective operation */
   MPID_THREAD_UNLOCK(comm->ADIctx,comm);

   return mpi_errno;
}   /* MPID_FN_Scatterv */
#endif   /* MPID_Scatterv */


/**********************************************************************/
#ifdef MPID_Allgather
int
MPID_FN_Allgather (void *sendbuf, int sendcnt,
                   struct MPIR_DATATYPE *sendtype, void *recvbuf,
                   int recvcnt, struct MPIR_DATATYPE *recvtype,
                   struct MPIR_COMMUNICATOR *comm)
{
   int mpi_errno = MPI_SUCCESS;
   int size, rank;

   /* Get my rank and the size of the communicator */
   (void) MPIR_Comm_size (comm, &size);
   (void) MPIR_Comm_rank (comm, &rank);
   /* Switch communicators to the hidden collective */
   comm = comm->comm_coll;

/* do something here! */

   return mpi_errno;
}   /* MPID_FN_Allgather */
#endif   /* MPID_Allgather */


/**********************************************************************/
#ifdef MPID_Allgatherv
int
MPID_FN_Allgatherv (void *sendbuf, int sendcnt,
                    struct MPIR_DATATYPE *sendtype, void *recvbuf,
                    int *recvcnts, int *displs,
                    struct MPIR_DATATYPE *recvtype,
                    struct MPIR_COMMUNICATOR *comm)
{
   int mpi_errno = MPI_SUCCESS;
   int size, rank;

   /* Get my rank and the size of the communicator */
   (void) MPIR_Comm_size (comm, &size);
   (void) MPIR_Comm_rank (comm, &rank);
   /* Switch communicators to the hidden collective */
   comm = comm->comm_coll;

/* do something here! */

   return mpi_errno;
}   /* MPID_FN_Allgatherv */
#endif   /* MPID_Allgatherv */


/**********************************************************************/
#ifdef MPID_Alltoall
int
MPID_FN_Alltoall (void *sendbuf, int sendcnt,
                  struct MPIR_DATATYPE *sendtype, void *recvbuf,
                  int recvcnt, struct MPIR_DATATYPE *recvtype,
                  struct MPIR_COMMUNICATOR *comm)
{
   int mpi_errno = MPI_SUCCESS;
   int size, rank;

   /* Get my rank and the size of the communicator */
   (void) MPIR_Comm_size (comm, &size);
   (void) MPIR_Comm_rank (comm, &rank);
   /* Switch communicators to the hidden collective */
   comm = comm->comm_coll;

   /* Lock for collective operation */
   MPID_THREAD_LOCK(comm->ADIctx, comm);

/* do something here! */

   /* Unlock for collective operation */
   MPID_THREAD_UNLOCK(comm->ADIctx,comm);

   return mpi_errno;
}   /* MPID_FN_Alltoall */
#endif   /* MPID_Alltoall */


/**********************************************************************/
#ifdef MPID_Alltoallv
int
MPID_FN_Alltoallv (void *sendbuf, int *sendcnts, int *sdispls,
                   struct MPIR_DATATYPE *sendtype, void *recvbuf,
                   int *recvcnts, int *rdispls,
                   struct MPIR_DATATYPE *recvtype,
                   struct MPIR_COMMUNICATOR *comm)
{
   int mpi_errno = MPI_SUCCESS;
   int size, rank;

   /* Get my rank and the size of the communicator */
   (void) MPIR_Comm_size (comm, &size);
   (void) MPIR_Comm_rank (comm, &rank);
   /* Switch communicators to the hidden collective */
   comm = comm->comm_coll;

   /* Lock for collective operation */
   MPID_THREAD_LOCK(comm->ADIctx, comm);

/* do something here! */

   /* Unlock for collective operation */
   MPID_THREAD_UNLOCK(comm->ADIctx,comm);

   return mpi_errno;
}   /* MPID_FN_Alltoallv */
#endif   /* MPID_Alltoallv */


/**********************************************************************/
#ifdef MPID_Reduce
int
MPID_FN_Reduce (void *sendbuf, void *recvbuf, int count,
                struct MPIR_DATATYPE *datatype, MPI_Op op, int root,
                struct MPIR_COMMUNICATOR *comm)
{
   int mpi_errno = MPI_SUCCESS;
   int size, rank, stride;
   struct MPIR_OP *op_ptr;
   MPI_User_function *uop;
   static char myname[] = "MPI_REDUCE";
   MPI_Aint lb, ub, extent;

   /* Is root within the communicator? */
   (void) MPIR_Comm_size (comm, &size);

#ifndef MPIR_NO_ERROR_CHECKING
   if ( root >= size )
      mpi_errno = MPIR_Err_setmsg (MPI_ERR_ROOT, MPIR_ERR_ROOT_TOOBIG, myname,
                                   (char *)0, (char *)0, root, size);
   else if ( root < 0 )
      mpi_errno = MPIR_Err_setmsg (MPI_ERR_ROOT, MPIR_ERR_DEFAULT, myname,
                                   (char *)0, (char *)0, root);
   if ( mpi_errno )
      return MPIR_ERROR(comm, mpi_errno, myname);
#endif

   /* See the overview in Collection Operations for why this is ok */
   if ( count == 0 ) return MPI_SUCCESS;

   /* Get my rank and switch communicators to the hidden collective */
   (void) MPIR_Comm_rank (comm, &rank);

   comm = comm->comm_coll;
   op_ptr = MPIR_GET_OP_PTR(op);
   MPIR_TEST_MPI_OP(op, op_ptr, comm, myname);
   uop = op_ptr->op;
   MPIR_Type_get_limits(datatype, &lb, &ub);
   extent = ub - lb;
   stride = extent * count;

   /* Lock for collective operation */
   MPID_THREAD_LOCK(comm->ADIctx,comm);

   /* MPI complete reference (Snir, Otto, Dongarra...), p177: advice to
    * implementors not respected.  The same result might NOT be obtained
    * whenever the function is applied on the same arguments appearing in
    * the same order: to take advantage of the physical location of the
    * processors, the results might vary with the root process (depending
    * on the architectures) */

   /* if the operation is not commutative, then it might be faster to
    * MPI_Gather to the root proc and let it compute everything.  That
    * also depends on the message size (size * count * extent).  If the
    * message size is very very large, it might be better ignore the
    * protocol levels and resort to an "hypercube algorithm":  for 8 procs:
    *  - phase 1: 1 sends {1} to 0 and 0 computes {0} * {1}
    *             5 sends {5} to 4 and 4 computes {4} * {5}
    *             3 sends {3} to 2 and 2 computes {2} * {3}
    *             7 sends {7} to 6 and 6 computes {6} * {7}
    *  - phase 2: 2 sends {2*3} to 0 and 0 computes {0*1} * {2*3}
    *             6 sends {6*7} to 4 and 4 computes {4*5} * {6*7}
    *  - phase 3: 4 sends {4*5*6*7} to 0 and 0 computes {0*1*2*3} * {4*5*6*7}.
    * But that could incur several WAN-TCP latencies in sequence...  That
    * decision must be made comparing the computation time (msg size) and
    * the latency. */
   if ( !op_ptr->commute )
   {
      void *tmp_buf;
      int i;

      if ( rank == root )
      {
         g_malloc_chk(tmp_buf, void *, size * stride);
         tmp_buf = (void *) ((char *) tmp_buf - lb);
      }   /* endif */
      mpi_errno = MPI_Gather(sendbuf, count, datatype->self, tmp_buf, count,
                             datatype->self, root, comm->self);
      if ( mpi_errno == MPI_SUCCESS  &&  rank == root )
      {
         /* MPI Standard: the operation is always assumed to be associative */
         /* copy the last data element into recvbuf */
         copy_buf((char *) tmp_buf + lb, size - 1,
                  (char *) recvbuf + lb, 0, stride);
         for (i = size-2; i >= 0; i--)
            (*uop)((char *) tmp_buf + i * stride, recvbuf, &count,
                   &(datatype->self));
      }   /* endif */
      if ( rank == root )
         g_free(tmp_buf);
   }
   else   /* commutative operation (and always assumed associative!) */
   {
      comm_hierarchy_t set_info;
      int set;
      void *tmp_buf;

      /* rename the clusters to have only zeros in root's column */
      update_cluster_ids(root, comm);

      /* find the sets of procs among which I will send/recv msgs */
      update_comm_sets(comm, &set_info);

      /* if I'm not the global root proc then I need to allocate a temporary
       * buffer to hold the intermediate value in the computation. */
      if ( rank != root )
      {
         g_malloc_chk(recvbuf, void *, stride);
         recvbuf = (void *) ((char *) recvbuf - lb);
      }   /* endif */
      /* copy my element into the recv buffer */
      copy_buf((char *)sendbuf + lb, 0, (char *) recvbuf + lb, 0,
               stride);

      /* allocate a temporary recv buffer */
      g_malloc_chk(tmp_buf, void *, stride);
      tmp_buf = (void *) ((char *) tmp_buf - lb);

      for (set = set_info.num-1; set >= 0; set--)
      {
         comm_set_t current_set = set_info.sets[set];

         /* for high latencies (WAN-TCP) it's more efficient to Gather
          * the data to the local root process and let it compute. */
         /* for very very large msg sizes or small latencies, use a
          * binomial-tree algorithm. */

         /* recving messages and computing a local intermediate value */
         /* sending combined messages */
         if ( current_set.level == MPICHX_WAN_LEVEL ) /* flat-tree algorithm */
            mpi_errno = flat_tree_reduce(current_set, recvbuf, extent, lb,
                                         count, datatype->self, comm->self,
                                         uop, tmp_buf);
         else   /* binomial-tree algo */
            mpi_errno = hypercube_reduce(current_set, recvbuf, count,
                                         datatype->self, comm->self, extent, lb,
                                         uop, tmp_buf);
         if ( mpi_errno ) break;
      }   /* endfor */

      g_free(tmp_buf);
      if ( rank != root )
         g_free(recvbuf);
   }   /* endif: end of commutative case */

   /* Unlock for collective operation */
   MPID_THREAD_UNLOCK(comm->ADIctx,comm);

   return mpi_errno;
}   /* MPID_FN_Reduce */
#endif   /* MPID_Reduce */


/**********************************************************************/
#ifdef MPID_Allreduce
int
MPID_FN_Allreduce (void *sendbuf, void *recvbuf, int count,
                   struct MPIR_DATATYPE *datatype, MPI_Op op,
                   struct MPIR_COMMUNICATOR *comm)
{
   int mpi_errno = MPI_SUCCESS;

   return mpi_errno;
}   /* MPID_FN_Allreduce */
#endif   /* MPID_Allreduce */


/**********************************************************************/
#ifdef MPID_Reduce_scatter
int
MPID_FN_Reduce_scatter (void *sendbuf, void *recvbuf, int *recvcnts,
                        struct MPIR_DATATYPE *datatype, MPI_Op op,
                        struct MPIR_COMMUNICATOR *comm)
{
   int mpi_errno = MPI_SUCCESS;
   int size, rank;
   static char myname[] = "MPI_REDUCE_SCATTER";

   /* Get my rank and the size of the communicator */
   (void) MPIR_Comm_size (comm, &size);
   (void) MPIR_Comm_rank (comm, &rank);
   /* Switch communicators to the hidden collective */
   comm = comm->comm_coll;

   return mpi_errno;
}   /* MPID_FN_Reduce_scatter */
#endif   /* MPID_Reduce_scatter */


/**********************************************************************/
#ifdef MPID_Scan
int
MPID_FN_Scan (void *sendbuf, void *recvbuf, int count,
              struct MPIR_DATATYPE *datatype, MPI_Op op,
              struct MPIR_COMMUNICATOR *comm)
{
   int mpi_errno = MPI_SUCCESS;
   int size, rank;

   /* See the overview in Collection Operations for why this is ok */
   if (count == 0) return MPI_SUCCESS;

   /* Get my rank and the size of the communicator */
   (void) MPIR_Comm_size (comm, &size);
   (void) MPIR_Comm_rank (comm, &rank);
   /* Switch communicators to the hidden collective */
   comm = comm->comm_coll;

   /* Lock for collective operation */
   MPID_THREAD_LOCK(comm->ADIctx,comm);

   /* Unlock for collective operation */
   MPID_THREAD_UNLOCK(comm->ADIctx,comm);

   return mpi_errno;
}   /* MPID_FN_Scan */
#endif   /* MPID_Scan */


